#!/usr/bin/perl

# $Id: make_group.pl,v 1.4 2012/10/18 20:24:41 peter Exp $

use strict;
use warnings;

use Fcntl;
use Errno;
use Net::LDAP;
use Getopt::Std;

# Requires:
# openldap24-client
# p5-perl-ldap

# option parsing:
# -r <randsleep> - sleep a random number of seconds to avoid hammering the server
# -n - dont do it, just leave /etc/group.new
# -q - quiet
# -v - verbose

sub usage {
	print STDERR "Usage: perl $0 [-r randsleep] [-q] [-v]\n";
	exit 1;
}

# declare the perl command line flags/options we want to allow
my %options=();
getopts("nqvr:", \%options) || usage();
usage() if $ARGV[0];

if (defined $options{v}) {
	if (defined ($options{r})) {
		print "random sleep time: $options{r}\n";
	}
	if (defined ($options{n})) {
		print "do not activate\n";
	}
}

if (defined($options{r})) {
	my $r = int (rand($options{r}) + 1.0);
	print "sleeping for $r\n" unless defined($options{q});
	sleep($r);
}

my $ldap = Net::LDAP->new(["ldap1.freebsd.org", "ldap2.freebsd.org", "ldap-master.freebsd.org"]) or die "Error in creating new Connection: $@";;
my $mesg = $ldap->search(filter=>"(&(objectClass=posixGroup)(cn=*))", base=>"ou=groups,dc=freebsd,dc=org");
$mesg->code && die $mesg->error;

my %add_gid = ();
my %add_names = ();

# ldap:*:389:389::0:0:OpenLDAP Server:/nonexistent:/sbin/nologin
while (my $entry=$mesg->pop_entry()) {
	my $members = join (",", sort $entry->get_value("memberUid"));

	my $gid = $entry->get_value("gidNumber") || die "ldap record has no gidNumber";
	my $name = $entry->get_value("cn") || die "ldap record has no cn";
	my $line = join (":", $entry->get_value("cn") || "teapot",
		"*" ,
		$entry->get_value("gidNumber") || 65533,
		$members || "");

	$add_gid{$gid} = $line;
	$add_names{$name} = $line;
}

$ldap->unbind;

sub write_entries {
	printf NGR "## LDAP BEGIN\n";
	foreach my $key (sort { $a <=> $b } keys %add_gid) {
		printf NGR "%s\n", $add_gid{$key};
	}
	printf NGR "## LDAP END\n";
}
my $state = 0;

if (!sysopen(MGR, "/etc/group", O_RDWR | O_EXLOCK | O_NONBLOCK)) {
	if ($!{EWOULDBLOCK}) {
		printf STDERR "group busy, aborting\n" unless defined $options{q};
		exit 0;
	}
	die "open /etc/group O_RDWR | O_EXLOCK | O_NONBLOCK: $!";
}
sysopen(NGR, "/etc/group.new", O_RDWR | O_CREAT | O_TRUNC, 0644) || die "open /etc/group.new O_RDWR | O_CREAT | O_TRUNC: $!";

while (<MGR>) {
	my ($name, $pw, $gid, $junk);
	chomp;
	my $line = $_;
	if (/^## LDAP BEGIN/) {
		$state = 1;
		next;
	}
	if (/^## LDAP END/) {
		$state = 2;
		next;
	}
	($name, $pw, $gid, $junk) = split(/:/, $line);
	if ($add_names{$name}) {
		# print "trimming duplicate name\n";
		next;
	}
	# if we found the "END" marker, write them here.
	if ($state == 2) {
		write_entries();
		$state = 3;
	}
	if ($state != 1) {
		printf NGR "%s\n", $line;
	}
}
# state 3 means we've alread written our lines.
if ($state != 3) {
	write_entries();
}

if (defined ($options{n})) {
	close(MGR);
	close(NGR);
	print "Tempfile left as /etc/group.new for review\n";
	exit 0;
}

rename("/etc/group.new", "/etc/group") || die "Could not rename /etc/group.new to /etc/group: $!";
close(MGR);
close(NGR);

