#!/usr/bin/perl

# $Id: make_k5login.pl,v 1.5 2012/10/18 20:24:41 peter Exp $

use strict;
use warnings;

use Fcntl;
use Errno;
use Net::LDAP;
use Getopt::Std;

# Requires:
# openldap24-client
# p5-perl-ldap

# option parsing:
# -r <randsleep> - sleep a random number of seconds to avoid hammering the server
# -o objClass - which ldap account class (posixAccount, adminAccount, freebsdAccount, etc)
# -n - dont do it, just leave /root/.k5login.new
# -q - quiet
# -v - verbose

sub usage {
	print STDERR "Usage: perl $0 [-r randsleep] [-o objclass] [-q] [-v]\n";
	exit 1;
}

# declare the perl command line flags/options we want to allow
my %options=();
getopts("nqvo:r:", \%options) || usage();
usage() if $ARGV[0];

my $objclass = "adminAccount";
if (defined $options{o}) {
	$objclass = $options{o};
}


if (defined $options{v}) {
	print "LDAP objclass: $objclass\n";
	if (defined ($options{r})) {
		print "random sleep time: $options{r}\n";
	}
	if (defined ($options{n})) {
		print "do not activate\n";
	}
}

if (defined($options{r})) {
	my $r = int (rand($options{r}) + 1.0);
	print "sleeping for $r\n" unless defined($options{q});
	sleep($r);
}

my %add_uid = ();
my %add_names = ();

my $ldap = Net::LDAP->new(["ldap1.freebsd.org", "ldap2.freebsd.org", "ldap-master.freebsd.org"]) or die "Error in creating new Connection: $@";;


foreach my $oc (split(/\+/, $objclass)) {
	my $mesg = $ldap->search(filter=>"(&(objectClass=$oc)(uid=*))", base=>"ou=users,dc=freebsd,dc=org");
	$mesg->code && die $mesg->error;


	# ldap:*:389:389::0:0:OpenLDAP Server:/nonexistent:/sbin/nologin
	while (my $entry=$mesg->pop_entry()) {
		my $uid = $entry->get_value("uidNumber") || die "ldap record has no uidNumber";
		my $name = $entry->get_value("uid") || die "ldap record has no uid";

		my $line = "$name/root\@FREEBSD.ORG";
		if ($name eq "root" || $uid == 0) {
			# How about we don't replace "root" or add uid = 0 lines..
			die "Refusing to add line $line";
		}
		$add_uid{$uid} = $line;
		$add_names{$line} = $line;
	}
}

$ldap->unbind;

sub write_entries {
	printf NKL "## LDAP BEGIN\n";
	foreach my $key (sort { $a <=> $b } keys %add_uid) {
		printf NKL "%s\n", $add_uid{$key};
	}
	printf NKL "## LDAP END\n";
}
my $state = 0;

if (!sysopen(OKL, "/root/.k5login", O_RDWR | O_EXLOCK | O_NONBLOCK)) {
	if ($!{EWOULDBLOCK}) {
		printf STDERR ".k5login busy, aborting\n" unless defined $options{q};
		exit 0;
	}
	die "open /root/.k5login O_RDWR | O_EXLOCK | O_NONBLOCK: $!";
}
sysopen(NKL, "/root/.k5login.new", O_RDWR | O_CREAT | O_TRUNC, 0600) || die "open /root/.k5login.new O_RDWR | O_CREAT | O_TRUNC: $!";

while (<OKL>) {
	my ($name, $pw, $uid, $junk);
	chomp;
	my $line = $_;
	if (/^## LDAP BEGIN/) {
		$state = 1;
		next;
	}
	if (/^## LDAP END/) {
		$state = 2;
		next;
	}
	if ($add_names{$line}) {
		# print "trimming duplicate name\n";
		next;
	}
	# if we found the "END" marker, write them here.
	if ($state == 2) {
		write_entries();
		$state = 3;
	}
	if ($state != 1) {
		printf NKL "%s\n", $line;
	}
}
# state 3 means we've alread written our lines.
if ($state != 3) {
	write_entries();
}

if (defined ($options{n})) {
	close(OKL);
	close(NKL);
	print "Tempfile left as /root/.k5login.new for review\n";
	exit 0;
}
rename("/root/.k5login.new", "/root/.k5login") || die "Could not rename /root/.k5login.new to /root/.k5login: $!";
close(OKL);
close(NKL);
