#!/usr/bin/perl

# $Id: make_masterpasswd.pl,v 1.15 2012/11/11 04:49:17 peter Exp $

use strict;
use warnings;

use Fcntl;
use Errno;
use Net::LDAP;
use Getopt::Std;

# Requires:
# openldap24-client
# p5-perl-ldap

# option parsing:
# -r <randsleep> - sleep a random number of seconds to avoid hammering the server
# -o objClass - which ldap account class (posixAccount, adminAccount, freebsdAccount, etc)
# -s shellattr - which shell to use (loginShell, adminShell)
# -d defaultshell - if shellattr isn't found in this record, use this instead (/sbin/nologin)
# -D fallbackshell - if shellattr is set but doesn't exist, use this instead (/bin/tcsh)
# -f forceshell - ignore shellattr, force this shell
# -n - dont do it, just leave /etc/master.passwd.new
# -q - quiet
# -v - verbose

sub usage {
	print STDERR "Usage: perl $0 [-r randsleep] [-o objclass] [-s shellattr] [-d defshell] [-f forceshell] [-D fallbacksell] [-q] [-v]\n";
	exit 1;
}

# declare the perl command line flags/options we want to allow
my %options=();
getopts("nqvo:r:s:d:D:f:", \%options) || usage();
usage() if $ARGV[0];

my $objclass = "freebsdAccount";
if (defined $options{o}) {
	$objclass = $options{o};
}

my $shellattr = "loginShell";
if (defined $options{s}) {
	$shellattr = $options{s};
}

my $forceshell;
if (defined $options{f}) {
	$forceshell = $options{f};
}

my $defaultshell;
if (defined $options{d}) {
	$defaultshell = $options{d};
}

my $fallbackshell;
if (defined $options{D}) {
	$fallbackshell = $options{d};
}


if (defined $options{v}) {
	print "LDAP objclass: $objclass\n";
	print "LDAP shellattr: $shellattr\n";
	if (defined ($options{r})) {
		print "random sleep time: $options{r}\n";
	}
	if (defined ($options{n})) {
		print "do not activate\n";
	}
}

if (defined($options{r})) {
	my $r = int (rand($options{r}) + 1.0);
	print "sleeping for $r\n" unless defined($options{q});
	sleep($r);
}

my %add_uid = ();
my %add_names = ();

my $ldap = Net::LDAP->new(["ldap1.freebsd.org", "ldap2.freebsd.org", "ldap-master.freebsd.org"]) or die "Error in creating new Connection: $@";;


foreach my $oc (split(/\+/, $objclass)) {
	#print "checking for $oc\n";
	my $mesg = $ldap->search(filter=>"(&(objectClass=$oc)(uid=*))", base=>"ou=users,dc=freebsd,dc=org");
	$mesg->code && die $mesg->error;


	# ldap:*:389:389::0:0:OpenLDAP Server:/nonexistent:/sbin/nologin
	while (my $entry=$mesg->pop_entry()) {
		my $uid = $entry->get_value("uidNumber") || die "ldap record has no uidNumber";
		my $name = $entry->get_value("uid") || die "ldap record has no uid";

		my $shell;
		foreach my $sh (split(/,/, $shellattr)) {
			if ($entry->exists($sh)) {
				$shell = $entry->get_value($sh);
				if (! -x "$shell" && defined($fallbackshell)) {
					$shell = $fallbackshell;
				}
				last;
			} else {
				$shell = $defaultshell;
			}
		}
		if (! defined($shell)) {
			$shell = "/sbin/nologin";
		}
		if ($forceshell) {
			$shell = $forceshell;
		}
		my $line = join (":", $entry->get_value("uid") || "teapot",
			"*",
			$entry->get_value("uidNumber") || 65533,
			$entry->get_value("gidNumber") || 65533,
			$entry->get_value("loginClass") || "",
			"0", 
			"0", 
			$entry->get_value("gecos") || $entry->get_value("cn") || "User &",
			$entry->get_value("homeDirectory") || "/nonexistent",
			$shell);
		if ($name eq "root" || $uid == 0) {
			# How about we don't replace "root" or add uid = 0 lines..
			die "Refusing to add line $line";
		}
		$add_uid{$uid} = $line;
		$add_names{$name} = $line;
	}

}
$ldap->unbind;

sub write_entries {
	printf NPW "## LDAP BEGIN\n";
	foreach my $key (sort { $a <=> $b } keys %add_uid) {
		printf NPW "%s\n", $add_uid{$key};
	}
	printf NPW "## LDAP END\n";
}
my $state = 0;

if (!sysopen(MPW, "/etc/master.passwd", O_RDWR | O_EXLOCK | O_NONBLOCK)) {
	if ($!{EWOULDBLOCK}) {
		printf STDERR "master.passwd busy, aborting\n" unless defined $options{q};
		exit 0;
	}
	die "open /etc/master.passwd O_RDWR | O_EXLOCK | O_NONBLOCK: $!";
}
sysopen(NPW, "/etc/master.passwd.new", O_RDWR | O_CREAT | O_TRUNC, 0600) || die "open /etc/master.passwd.new O_RDWR | O_CREAT | O_TRUNC: $!";

while (<MPW>) {
	my ($name, $pw, $uid, $junk);
	chomp;
	my $line = $_;
	if (/^## LDAP BEGIN/) {
		$state = 1;
		next;
	}
	if (/^## LDAP END/) {
		$state = 2;
		next;
	}
	($name, $pw, $uid, $junk) = split(/:/, $line);
	if ($add_names{$name}) {
		# print "trimming duplicate name\n";
		next;
	}
	# if we found the "END" marker, write them here.
	if ($state == 2) {
		write_entries();
		$state = 3;
	}
	if ($state != 1) {
		printf NPW "%s\n", $line;
	}
}
# state 3 means we've alread written our lines.
if ($state != 3) {
	write_entries();
}

if (defined ($options{n})) {
	close(MPW);
	close(NPW);
	print "Tempfile left as /etc/master.passwd.new for review\n";
	exit 0;
}

exec '/usr/sbin/pwd_mkdb', '-p', '/etc/master.passwd.new';
die "Could not execve /usr/sbin/pwd_mkdb: $!";
