#!/usr/bin/perl

# $Id: make_sshkeys.pl,v 1.7 2012/10/18 20:24:41 peter Exp $

# LDAP is searched and we store the server's idea of the keylist in $ldap_keys{}
# We scan the keystore directory and store that in $user_keys{}
# Then compare/add/update/remove.

# Requires:
# openldap24-client
# p5-perl-ldap

# option parsing:
# -r <randsleep> - sleep a random number of seconds to avoid hammering the server
# -d keydir - where the keys are stored on this machine
# -k keyValue - which ldap attribute to export (sshPublicKey, adminPublicKey, etc)
# -o objClass - which ldap account class (posixAccount, adminAccount, freebsdAccount, etc)
# -q - quiet
# -v - verbose

use strict;
use warnings;

use Net::LDAP;
use Getopt::Std;

sub usage {
	print STDERR "Usage: perl $0 [-r randsleep] [-d keydir] [-k keyattr] [-o objclass] [-q] [-v]\n";
	exit 1;
}

# declare the perl command line flags/options we want to allow
my %options=();
getopts("qvr:d:k:o:", \%options) || usage();
usage() if $ARGV[0];

my $keydir = "/etc/ssh-keys";
if (defined $options{d}) {
	$keydir = $options{d};
}
# Get rid of any symlinks to dumpster
unlink($keydir);
mkdir($keydir);

my $keyval = "sshPublicKey";
if (defined $options{k}) {
	$keyval = $options{k};
}

my $objclass = "freebsdAccount";
if (defined $options{o}) {
	$objclass = $options{o};
}

if (defined $options{v}) {
	print "Key directory: $keydir\n";
	print "LDAP attribute: $keyval\n";
	print "LDAP objclass: $objclass\n";
	if (defined ($options{r})) {
		print "random sleep time: $options{r}\n";
	}
}

if (defined($options{r})) {
	my $r = int (rand($options{r}) + 1.0);
	print "sleeping for $r\n" unless defined($options{q});
	sleep($r);
}

# Compare arrays to see if they have the same contents, regardless of order
sub identical {
	my ($left, $right) = @_;
	return 0 if scalar @$left != scalar @$right;
	my %hash;
	@hash{@$left, @$right} = ();
	return scalar keys %hash == scalar @$left;
}
sub unique {
	my %hashTemp = map { $_ => 1 } @_;
	my @array_out = sort keys %hashTemp;
	return @array_out;
}

my %ldap_keys = ();
my %user_keys = ();


# Search and parse LDAP

my $ldap = Net::LDAP->new(["ldap2.freebsd.org", "ldap1.freebsd.org", "ldap-master.freebsd.org"]) or die "Error in creating new Connection: $@";

foreach my $oc (split(/\+/, $objclass)) {

	my $mesg = $ldap->search(filter=>"(&(objectClass=$oc)(uid=*))", base=>"ou=users,dc=freebsd,dc=org");
	$mesg->code && die $mesg->error;

	while (my $entry=$mesg->pop_entry()) {
		my $uid = $entry->get_value("uid");
		#@{$ldap_keys{$uid}} = $entry->get_value("$keyval");
		foreach my $attr (split(/([+,])/, $keyval)) {
			#print "attr = $attr\n";
			last if ($attr eq ',' && $ldap_keys{$uid});
			next if ($attr eq "+" || $attr eq ",");
			#print "searching for $attr for $uid\n";
			if ($entry->exists("$attr")) {
				#print "adding $attr for $uid\n";
				push @{$ldap_keys{$uid}}, $entry->get_value($attr);
			}
		}
	}
}

$ldap->unbind;


# Scan the existing keys

opendir(KEYS, $keydir) or die "opendir $keydir: $!";
while (my $file = readdir(KEYS)) {
	# skip any dot files
	next if ($file =~ m/^\./);
	# We only want files
	next unless (-f "$keydir/$file");

	open(KEYFILE, "< $keydir/$file") || die "read $keydir/$file: $!";
	while (<KEYFILE>) {
		chomp;
		push (@{$user_keys{$file}}, $_);
	}
	close(KEYFILE);
	@{$user_keys{$file}} = unique(@{$user_keys{$file}});
}
closedir(KEYS);

# For each key that we found on disk, if it is identical to the server's copy, delete it from
# our server's copy.
# If it is no longer on the server, delete it here.

foreach my $uid (sort keys %ldap_keys) {
	if ($ldap_keys{$uid}) {
		push @{$ldap_keys{$uid}}, "## DO NOT EDIT - auto generated - http://wiki.freebsd.org/clusteradm/ssh-keys";
		@{$ldap_keys{$uid}} = unique(@{$ldap_keys{$uid}});
	}
}
foreach my $uid (sort keys %user_keys) {
	if (! $ldap_keys{$uid}) {
		print "Deleting key for $uid\n" unless defined($options{q});
		unlink("$keydir/$uid") || die "unlink $keydir/$uid: $!";
		next;
	}
	if (identical(\@{$ldap_keys{$uid}}, \@{$user_keys{$uid}})) {
		#print "Key for $uid is identical\n" unless defined($options{q});
		delete $ldap_keys{$uid};
	} else {
		print "Key $uid needs updating\n" unless defined($options{q});
	}
}

# Anything left after the sweep above either needs to be updated or installed as a new key
foreach my $uid (sort keys %ldap_keys) {
	if (! $user_keys{$uid}) {
		print "Adding new key for $uid\n" unless defined($options{q});
	} else {
		print "Updating key for $uid\n" unless defined($options{q});
	}
	open(NEWKEY, "> $keydir/$uid.new") || die "read $keydir/$uid.new: $!";
	foreach my $key (sort @{$ldap_keys{$uid}}) {
		printf NEWKEY "%s\n", $key || die "print to newkey file $keydir/$uid.new: $!";
	}
	close NEWKEY || die "close $keydir/$uid.new: $!";
	rename ("$keydir/$uid.new", "$keydir/$uid") || die "rename $keydir/$uid.new - $keydir/$uid: $!";
}
